/* This is a "base" driver for USB.
 * It is not as comprehensive as my other "base" drivers, becase
 * USB devices can be a lot different from each other.
 * If you want in-depth examples, download some of my other USB drivers
 * from
 *  http://www.bolthole.com/solaris/drivers/
 *
 * What this driver does is attach to any USB ID you specify,
 * and print out endpoint information.
 *
 * BEFORE INSTALLING THE DRIVER, rename everything from 'usbbbase' to
 * 'yourdrivername'. This includes stuff inside this file, and
 * in the Makefile
 *
 *    Philip Brown,  Feb 2002
 *
 *  @(#) usbbase.c 1.11@(#)
 */

#define _SYS_VARARGS_H         /* sys/varargs.h hack for gcc2 */
#include <stdarg.h>

#include <sys/usb/usba.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/stream.h>
#include <stropts.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


#define USBBASE_BUFFER_SIZE 512
#define USBBASE_MAX_TRANSFER 128


/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"usbbase"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	usbbase_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	usbbase_probe(dev_info_t *dip);
static	int	usbbase_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	usbbase_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	usbbase_print(dev_t, char*);

static	int	usbbase_read(dev_t dev, struct uio* ustr, cred_t * credp);

static	int	usbbase_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	usbbase_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	usbbase_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	usbbase_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	usbbase_cb_ops = {
	usbbase_open,		/* open() */
	usbbase_close,		/* close() */
	nodev,			/* strategy()		*/
	usbbase_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	usbbase_read,		/* read() */
	usbbase_write,		/* write() */
	usbbase_ioctl,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP|D_64BIT,	/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops usbbase_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	usbbase_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	usbbase_probe,		/* devo_probe */
	usbbase_attach,		/* devo_attach */
	usbbase_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&usbbase_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"USB test driver",
	&usbbase_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   usbbase_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * usbbase_glob_soft_handle;

/* This gets filled out on a per-instance basis by usbbase_attach */
struct usbbase_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex;
	kcondvar_t read_cv;
	
	int openflags; /* tell if we have been open()ed or not */

	usb_config_descr_t usb_config;
	int usb_interfacenum;
	usb_endpoint_descr_t endpoint;
	usb_interface_descr_t usb_interface_d;
	usb_pipe_handle_t usb_pipe_handle;

	/* since we are not a streams driver we ahve to make our
	 * own buffer/queue
	 * "end" is where to add new data. "start" is where read()
	 * should take data from
	 */
	unsigned char *readbuffer;
	unsigned char *readbuf_start,*readbuf_end;

};



/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
usbbase_debugprint(const char *format, ...)
{
	va_list args;
	va_start(args, format);

	if(debug_level==0){
		return;
	}

	/* DO NOT USE vcmn_err here!!! it will not be interrupt-safe.
	 * See the "CONTEXT" part of the manpage
	 */

	vprintf(format, args);
	va_end(args);
}



/*Like the usbbase_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See usbbase_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	cmn_err(CE_NOTE,"usbbase: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&usbbase_global_mutex, "usbbase: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct usbbase_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&usbbase_glob_soft_handle,
				sizeof(struct usbbase_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		cmn_err(CE_WARN,"soft_state_init failed");
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&usbbase_global_mutex);
		ddi_soft_state_fini(&usbbase_glob_soft_handle);
		cmn_err(CE_WARN, "usbbase: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE,"usbbase: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	usbbase_debugprint("usbbase: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	usbbase_debugprint("usbbase: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&usbbase_glob_soft_handle);
		mutex_destroy(&usbbase_global_mutex);
	}
	return (ret);
}



/**********************************************************************/
/*  Our read buffer handling section                                  */
/*  Depending on the type of your driver, you may not NEED this       */
/* But I figure you might like a pre-written one anyway.              */
/**********************************************************************/

/* return 0 on okay, 1 on fail.
 * handle whatever allocation the read queue stuff needs
 */
static int usbbase_readqueue_init(struct usbbase_softstate *softstate)
{
	softstate->readbuffer=kmem_zalloc(USBBASE_BUFFER_SIZE,KM_SLEEP);
	return 0;
}

static void usbbase_readqueue_finish(struct usbbase_softstate *softstate)
{
	kmem_free(softstate->readbuffer,USBBASE_BUFFER_SIZE);
}

/* Returns number of bytes in the read buffer.
 * Dont expect a sensible return unless you are holding the
 * instance_mutex.
 */
static int read_buffer_bytes(struct usbbase_softstate *softstate){
	int datalen;


	if(softstate->readbuf_start > softstate->readbuf_end){
		datalen=(&softstate->readbuffer[USBBASE_BUFFER_SIZE]) -
			softstate->readbuf_start;

		datalen+= softstate->readbuf_end - softstate->readbuffer;
	} else {
		datalen=softstate->readbuf_end - softstate->readbuf_start;
	}
	

	return datalen;
}

/*
 * Note that we have a circular buffer, and we do NOT overwrite.
 * After buffer is full, we drop further packets.
 *
 * YOU MUST HAVE instance_mutex BEFORE CALLING!!!
 */
static void
add_to_read_buffer(struct usbbase_softstate *softstate, char *data, int len){
	unsigned char *bufstart, *bufend;
	unsigned char *bufdatastart,*bufdataend;
	int copylen, bytesfree;

	if(softstate==NULL) {
		usbbase_debugprint("add_to_read_buffer: softstate==null");
		return;
	}
	if(data==NULL) {
		usbbase_debugprint("add_to_read_buffer: data==null");
		return;
	}
	bufstart=softstate->readbuffer;
	bufend=&softstate->readbuffer[USBBASE_BUFFER_SIZE];
	bufdatastart=softstate->readbuf_start;
	bufdataend=softstate->readbuf_end;


	bytesfree=USBBASE_BUFFER_SIZE - read_buffer_bytes(softstate);
	if(bytesfree==0){
		usbbase_debugprint("add_to_read_buffer: buffer full\n");
		/* silently discard */
		return;
	}
	if(len>bytesfree){
		copylen=bytesfree;
		len=bytesfree;
		usbbase_debugprint("Note: add_to_read_buffer: partial add\n");
	} else {
		copylen=len;
	}

	if(&bufdataend[copylen]  >= &softstate->readbuffer[USBBASE_BUFFER_SIZE]){
		copylen=(&softstate->readbuffer[USBBASE_BUFFER_SIZE]) -
			softstate->readbuf_end;
	}
	bcopy(data,bufdataend, copylen);
	bufdataend += copylen;
	if(bufdataend==bufend){
		bufdataend=bufstart;
	}
	len-= copylen;
	if(len>0){
		/* should only be here if we had to wrap around end of buffer*/
		copylen=len;
		bcopy(data,bufdataend, copylen);
		bufdataend += copylen;
		len -= copylen;
	}
	
	softstate->readbuf_end=bufdataend;
	
	if(len!=0){
		cmn_err(CE_WARN, "usbbase:add_to_read_buffer: internal error\n");
	}

} /* end of add_to_read_buffer() */

/* Make sure you have the instance_mutex before calling this.
 * Returns internal pointers to "clear" the buffer.
 */
static void read_buffer_reset(struct usbbase_softstate *softstate){
	softstate->readbuf_start=softstate->readbuffer;
	softstate->readbuf_end=softstate->readbuffer;
}

/************************************************************/
/*              End of readbuffer section                   */
/************************************************************/

/************************************************************/
/*       Start of callbacks section                         */
/*      You will probably only need EITHER the isoc ones,   */
/*      OR the other ones                                   */
/************************************************************/


/*
 * This is probably interrupt context. 
 */
static int
usbbase_isoc_callback(usb_pipe_handle_t pipe,
                        usb_opaque_t arg,
			usb_isoc_request_t *isoc_req)
{
	struct usbbase_softstate *softstate;
	unsigned char*chardata;
	int datalen;
	softstate=(struct usbbase_softstate*)arg;

	if(debug_level>1){
		usbbase_debugprint("usbbase_pipe_callback called\n");
	}

	if(isoc_req==NULL){
		printf("isoq_req==NULL\n");
		return (USB_SUCCESS);
	}


	printf("DEBUG: isoq_req = 0x%p\n",isoc_req);

#ifdef NEED_DEALLOC
	usb_deallocate_isoc_req(pipe, isoq_req);
#endif
	if(isoc_req->isoc_data==NULL){
		printf("DEBUG: isoc_data == NULL?!!\n");
	} else {
		mblk_t *mblkptr=isoc_req->isoc_data;
		datalen=mblkptr->b_wptr - mblkptr->b_rptr;
		printf("theoretical len of data is %d\n",datalen);
		
		freemsg(isoc_req->isoc_data);
	}

#ifdef OLDSTUFF
	/* If there is data, it is usually in 5-byte blocks */
	datalen=mblkptr->b_wptr - mblkptr->b_rptr;
	/*	printf("theoretical len of data is %d\n",datalen);*/
	chardata=mblkptr->b_rptr;

	mutex_enter(&softstate->instance_mutex);
	add_to_read_buffer(softstate,chardata, datalen);

	freemsg(mblkptr);
	cv_signal(&softstate->read_cv);

	mutex_exit(&softstate->instance_mutex);
#endif /* OLDSTUFF */

	return (USB_SUCCESS);
}
/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static int
usbbase_isoc_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason, usb_isoc_request_t *isoc_req,
			 uint_t flag)
{
	/*struct usbbase_softstate *softstate=(struct usbbase_softstate*)arg;*/

	usbbase_debugprint("usbbase_isoc_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	/* See USB_CC_xxxx defines to interpret complete_reason */

	if(complete_reason==9){
		usbbase_debugprint("You may need to set USB_FLAGS_SHORT_XFER_OK\n");
	}
	return (USB_SUCCESS);
}


/* This is probably interrupt context. 
 * mblk_t is a STREAMS style Message Block. Arrg.
 */
static int
usbbase_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	struct usbbase_softstate *softstate=(struct usbbase_softstate*)arg;
	unsigned char*chardata;
	int datalen;

	if(debug_level>1){
		usbbase_debugprint("usbbase_pipe_callback called\n");
	}

	if(mblkptr==NULL){
		printf("mblk==NULL\n");
		return (USB_SUCCESS);
	}

	mutex_enter(&softstate->instance_mutex);

	/* If there is data, it is usually in 5-byte blocks */
	datalen=mblkptr->b_wptr - mblkptr->b_rptr;
	/*	printf("theoretical len of data is %d\n",datalen);*/
	chardata=mblkptr->b_rptr;

	add_to_read_buffer(softstate,chardata, datalen);

	freemsg(mblkptr);
	cv_signal(&softstate->read_cv);
	mutex_exit(&softstate->instance_mutex);

	return (USB_SUCCESS);
}

/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static int
usbbase_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{
	/*struct usbbase_softstate *softstate=(struct usbbase_softstate*)arg;*/

	usbbase_debugprint("usbbase_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	/* See USB_CC_xxxx defines to interpret complete_reason */

	if(complete_reason==9){
		usbbase_debugprint("You may need to set USB_FLAGS_SHORT_XFER_OK\n");

		/* USB_FLAGS_SHORT_XFER_OK should be set in either
		 * usb_pipe_start_polling(), or usb_pipe_receive_bulk_data(),
		 * depending on whether you are doing regular data receive,
		 * or bulk data
		 */
		
	}
	return (USB_SUCCESS);
}


/**********************************************************************/
/**********************************************************************/


/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS to use.
 */
static int usbbase_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int usbbase_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct usbbase_softstate *softstate;
	int returnval=DDI_FAILURE;
	dev_t devt=(dev_t)arg;
	int instance=DEVT_TO_INST(devt);
	
	usbbase_debugprint("usbbase: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(usbbase_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static int usbbase_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	usbbase_debugprint("usbbase: probe called\n");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
	usbbase_debugprint("probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}

/* This routine can be called by usbbase_attach (if errors occur)
 * and usbbase_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void usbbase_cleanup(struct usbbase_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: usbbase_cleanup called with NULL");
		return;
	}
	usbbase_readqueue_finish(softstate);
   	instance=ddi_get_instance(softstate->dip);

	ddi_soft_state_free(usbbase_glob_soft_handle,instance);
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in usbbase_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  usbbase_detach()
 */
static int usbbase_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance,tmpint;
	struct usbbase_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;
	usb_endpoint_descr_t endpoint;

	cmn_err(CE_NOTE,"usbbase: attach called\n");

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	usbbase_debugprint("usbbase: usbbase_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "usbbase: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "usbbase: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "usbbase: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	usbbase_debugprint("usbbase: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(usbbase_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "usbbase: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(usbbase_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "usbbase: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(usbbase_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	
	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		usbbase_cleanup(softstate);
		return DDI_FAILURE;
	}


	usbbase_debugprint("USB config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	usbbase_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		usbbase_debugprint("  (resetting interface num to 0)\n");
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		usbbase_cleanup(softstate);
		return DDI_FAILURE;
	}

	usbbase_debugprint("number of endpoints = %d\n",
			 softstate->usb_interface_d.bNumEndpoints);

	
	for(tmpint=0;tmpint<softstate->usb_interface_d.bNumEndpoints; tmpint++){
		if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					tmpint,/*endpoint index num*/
					&endpoint,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
		{
			cmn_err(CE_WARN, "could not parse endpoint %d",
				tmpint);
			continue;
		}

		/* Note: endpoint.bEndpointAddress&USB_EPT_ADDR_MASK
		 *    is NOT always the same as the endpoint INDEX.
		 *    Since the index is what we care about at this level,
		 *    just print out the indexnum instead
		 */
		usbbase_debugprint("endpoint index=%d: direction = %s, attribs=%d\n",
		       tmpint,
		       (endpoint.bEndpointAddress&USB_EPT_DIR_IN)>0 ?
			   "IN" : "OUT",
		       endpoint.bmAttributes);
		switch(endpoint.bmAttributes & USB_EPT_ATTR_MASK){
			case USB_EPT_ATTR_CONTROL:
			usbbase_debugprint("endpoint is a CONTROL endpoint\n");
			break;
			case USB_EPT_ATTR_ISOCH:
			usbbase_debugprint("endpoint is an isochronous endpoint\n");
			break;
			case USB_EPT_ATTR_BULK:
			usbbase_debugprint("endpoint is a bulk transfer endpoint\n");
			break;
			case USB_EPT_ATTR_INTR:
			usbbase_debugprint("endpoint is an interrupt endpoint\n");
			break;

		}
		
	printf(" Endpoint %d: %x %x %x %x %x %x\n",
	       tmpint,
	       endpoint.bLength,
	       endpoint.bDescriptorType,
	       endpoint.bEndpointAddress,
	       endpoint.bmAttributes,
	       endpoint.wMaxPacketSize,
	       endpoint.bInterval);
		
	}


	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);
	cv_init(&softstate->read_cv, NULL, CV_DRIVER,NULL);

	if(usbbase_readqueue_init(softstate)!=0){
		return DDI_FAILURE;
	}

	/* It is often good practice to put the instance number
	 * in the nodename. However, since all USB devices
	 * get their own special slot-based name, that isnt
	 * strictly neccesary here
	 */
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);



	/* This little call all by itself, will put a whole bunch of
	 * useful information about the device in
	 * /var/adm/messages, aka "dmesg |tail"
	 */
	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

}


/* return 0 on success, or errno */
static int usbbase_read(dev_t dev, struct uio* ustr, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	struct usbbase_softstate *softstate;

	usbbase_debugprint("usbbase: read called\n");

	softstate=ddi_get_soft_state(usbbase_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"usbbase_read failed: cannot get softstate?");
		return DDI_FAILURE;
	}
   	return (ENXIO);

}

static int usbbase_write(dev_t dev, struct uio *uiop, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct usbbase_softstate *softstate;

   	usbbase_debugprint("usbbase: write called\n");

	softstate=ddi_get_soft_state(usbbase_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"usbbase_write failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	return (ENXIO);
}


#define POLLING_DEVICE

static int usbbase_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct usbbase_softstate *softstate;
#ifdef POLLING_DEVICE
	usb_pipe_policy_t policy,*policyptr;
#endif

	usbbase_debugprint("usbbase: open for instance %d\n", instance);

	softstate=ddi_get_soft_state(usbbase_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

/* If you get 'intr' or 'isoc' data from your device, you will want
 * this section here.
 * It is currently set up for an "isoc" type device.
 * you will need to adjust, and write your own callback routines,
 * as required.
 */
#ifdef POLLING_DEVICE
	policyptr=&policy;
	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	policyptr->pp_callback_arg=(void*)softstate;/* same for ALL callbacks*/

	policyptr->pp_isoc_callback=usbbase_isoc_callback;
	policyptr->pp_isoc_exception_callback=usbbase_isoc_exception_callback;

	/* we do not set these, because there are no regular endpoints */
	/*
	policyptr->pp_callback=usbbase_pipe_callback;
	policyptr->pp_exception_callback=usbbase_exception_callback;
	*/

#error   You must set pp_periodic_max_transfer_size according to the endpoint

	/* set this according to the specific endpoint you use */
	/* You can fake it by setting it to some random number, but
	 * note that SOME devices will get very unhappy if it is set too small,
	 * OR TOO LARGE!!
	 */
	policyptr->pp_periodic_max_transfer_size=endpoint.wMaxPacketSize;

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=0){
		usbbase_debugprint("usbbase: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}
	softstate->openflags=1;
	mutex_exit(&softstate->instance_mutex);

	read_buffer_reset(softstate);


	if(usb_pipe_open(softstate->dip,&softstate->endpoint, policyptr,
			 USB_FLAGS_SLEEP,
			 &softstate->usb_pipe_handle) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open failed");
		softstate->openflags=0;
		return (ENXIO);
	}

	if(usb_pipe_start_polling(softstate->usb_pipe_handle,
				  USB_FLAGS_SLEEP,
				  NULL, /* callback routine. ugh */
				  NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling of pipe1 FAILED");
		softstate->openflags=0;
		return ENXIO;
	}

#endif /* POLLING_DEVICE */

   	return 0;
}

static int usbbase_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct usbbase_softstate *softstate;


	softstate=ddi_get_soft_state(usbbase_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"usbbase_close failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	usbbase_debugprint("usbbase: close called\n");

/* See comments in usbbase_open() */
#ifdef POLLING_DEVICE
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=1){
		cmn_err(CE_WARN,"ERROR: usbbase_close: pipe not open");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
		
	}
	softstate->openflags=2;
	mutex_exit(&softstate->instance_mutex);

	usb_pipe_stop_polling(softstate->usb_pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);

	if(usb_pipe_close(&softstate->usb_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		usbbase_debugprint("ERROR: usb_pipe_close failed?");
		return ENXIO;
	}
	softstate->usb_pipe_handle=NULL;

	/* Okay, a little redundant since we do this in open(), but
	 * thats okay. It makes sure read_buffer_bytes returns 0.
	 */
	read_buffer_reset(softstate);
	softstate->openflags=0;
	
#endif /* POLLING_DEVICE */
   	return (0);
}

static int usbbase_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp){
/*
	int instance=DEVT_TO_INST(dev);
	struct usbbase_softstate *softstate;
*/

	usbbase_debugprint("usbbase: ioctl called\n");
   	return (ENXIO);
}

static int usbbase_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct usbbase_softstate *softstate;
   	instance=ddi_get_instance(dip);

	usbbase_debugprint("usbbase: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(usbbase_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"usbbase_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	usbbase_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}
